# ../scripts/custom/gg_winner_menu/gg_winner_menu.py

# ============================================================================
# >> IMPORTS
# ============================================================================
# Python Imports
#   ConfigObj
from configobj import ConfigObj
#   Path
from path import path
#   Random
from random import choice as rchoice
from random import sample
from random import shuffle

# Eventscripts Imports
#   ES
from es import getplayercount
from es import getplayername
from es import getplayerteam
from es import getuserid
from es import getUseridList
from es import isbot
from es import ServerCommand
from es import ServerVar
#   Gamethread
from gamethread import delayedname
from gamethread import cancelDelayed
#   Popuplib
from popuplib import delete
from popuplib import easymenu
from popuplib import exists
from popuplib import find

# GunGame Imports
#   Addons
from gungame51.core.addons.shortcuts import AddonInfo
#   Messaging
from gungame51.core.messaging.shortcuts import langstring
from gungame51.core.messaging.shortcuts import msg
# ============================================================================
# >> ADDON REGISTRATION/INFORMATION
# ============================================================================
info = AddonInfo()
info.name = 'gg_winner_menu'
info.title = 'GG Winner Menu'
info.author = 'Satoon101'
info.version = '1.6'
info.translations = ['gg_winner_menu']

# =============================================================================
# >> PUBLIC VARIABLE
# =============================================================================
ServerVar(info.name + '_version', info.version,
        info.title + ' Version').set(info.version)
ServerVar(info.name + '_version').makepublic()

# ============================================================================
# >> GLOBAL VARIABLES
# ============================================================================
# Store the ServerVar instances
allow_bots = ServerVar('gg_winner_menu_allow_bots')
team_most_kills = ServerVar('gg_winner_menu_team_most_kills')


# ============================================================================
# >> CLASSES
# ============================================================================
class _TeamPlayers(dict):
    def __getitem__(self, userid):
        userid = int(userid)
        if userid in self:
            return super(_TeamPlayers, self).__getitem__(userid)
        value = self[userid] = 0
        return value


class _TeamKillers(dict):
    def __init__(self):
        for team in (2, 3):
            self[team] = _TeamPlayers()

    def clear(self):
        for team in self:
            self[team].clear()


class _WinnerMenu(object):
    '''Class used to manage the Winner Menu'''

    def __init__(self):
        '''Called when the class is first initialized'''

        # Store the gametypes
        self.gametypes = ConfigObj(
            path(__file__).parent.joinpath('gg_winner_menu_gametypes.ini'))

        # Create all of the class attributes
        self.gametype = None
        self.specials = {}
        self.team_killers = _TeamKillers()
        self.winner = None
        self.multi = 0
        self.remainder = 0

    def match_reset(self):
        '''Resets all of the match stored attributes'''

        # Reset all of the match stored attributes
        self.team_killers.clear()

        # Reset all of the round stored attributes
        self.round_reset()

    def round_reset(self):
        '''Resets all of the round stored attributes'''

        # Reset all of the round stored attributes
        self.specials.clear()
        self.winner = None

    def load_gametype(self, gametype=None):
        '''Loads a gametype and stores one if given'''

        # Was a gametype given?
        if not gametype is None:

            # Is this a valid gametype?
            if gametype in self.gametypes:

                # Set the gametype as the current
                self.gametype = gametype

        # Is there a current gametype?
        if not self.gametype is None:

            # Load the current gametype
            ServerCommand(self.gametypes[self.gametype]['load'])

    def unload_gametype(self):
        '''Unloads the current gametype'''

        # Is there a current gametype?
        if not self.gametype is None:

            # Unload the gametype
            ServerCommand(self.gametypes[self.gametype]['unload'])

    def round_end(self, team, reason):
        '''Stores the winner on round_end'''

        # Get the round's winner
        self.winner = self.get_team_winner(team, reason)

    def get_team_winner(self, team, reason=None):
        '''Returns the winner of the round'''

        # Is the game commencing?
        if reason == 15:

            # If so, simply return None
            return None

        # Was there a "special" winner?
        if reason in self.specials:

            # Does the userid still exist on the server?
            if getuserid(self.specials[reason]):

                # Return the userid
                return self.specials[reason]

        # Get the player with the "most" kills on the winning team?
        if int(team_most_kills):

            # Did a team win?
            if team in self.team_killers:

                # Does the team have any killers?
                if self.team_killers[team]:

                    # Get the team's killers
                    killers = self.team_killers[team]

                    # Sort the team's killers
                    killers = sorted(killers,
                        key=lambda userid: killers[userid],
                        reverse=True)

                    # Return the player with the most kills on the winning team
                    return killers[0]

        # Get the "last" killer on the winning team?
        else:

            # Did the winning team have any kills?
            if team in self.specials:

                # Is the last killer still on the server?
                if getuserid(self.specials[team]):

                    # Return the last killer on the winning team
                    return self.specials[team]

        # Was there a winning team?
        if team in (2, 3):

            # Return a random player from the winning team
            return self.get_random_player(team)

        # Return any random player
        return self.get_random_player()

    def send_winner_menu(self, team=None):
        '''Sends the winner menu to the winner'''

        # Is this a "team" win?
        if team:

            # Is there a winner?
            if self.winner is None:

                # If not, get the team winner
                self.winner = self.get_team_winner(team)

        # Is the winner a bot?
        if isbot(self.winner):

            # Choose a random gametype if a bot wins?
            if int(allow_bots):

                # Choose a random gametype
                self.choose_random_gametype(self.winner)

                # No menu needs sent, so return
                return

            # Choose a random player from the Bots team
            self.winner = self.get_random_player(getplayerteam(self.winner))

        # Does the winner still exist on the server?
        if not getuserid(self.winner):

            # If not, get a random player
            self.winner = self.get_random_player(team)

        # Does the menu need sent?
        if self.winner is None:

            # If not, choose a random gametype
            self.choose_random_gametype()

            # No menu needs sent, so return
            return

        # Does menu exist?
        if exists('gg_winner_menu'):

            # Delete the menu
            delete('gg_winner_menu')

        # Create the easy menu
        gg_winner_menu = easymenu('gg_winner_menu', 'choice', self.chosen_type)

        # Set the menu's title
        gg_winner_menu.settitle(langstring(
            'Winner Menu Title', userid=self.winner))

        # Get the GameTypes to be chosen from
        types_to_choose = self.get_gametypes()

        # Loop through GameTypes
        for gametype in types_to_choose:

            # Add each option to the menu
            gg_winner_menu.addoption(gametype, gametype)

        # Get the multiples of 4 and the remainder
        self.multi, self.remainder = divmod(int(ServerVar('mp_chattime')), 4)

        # Loop sending menu
        self.menu_loop()

    def menu_loop(self):
        '''Loops and re-sends the menu until the time runs out'''

        # Set timeout in case we don't need to send popup
        timeout = 0

        # Is there still 4 seconds?
        if self.multi:

            # Decrement multi
            self.multi -= 1

            # Set timeout to 4
            timeout = 4

        # Is there a remainder?
        elif self.remainder:

            # Set timeout to remainder amount
            timout = self.remainder

        # Do we need to send the menu?
        if timeout:

            gg_winner_menu = find('gg_winner_menu')

            # Set the Timout amount
            gg_winner_menu.timeout('view', timeout)

            # Send the menu to the winner
            gg_winner_menu.send(self.winner)

            # Loop sending menu
            delayedname(timeout, 'gg_winner_menu', self.menu_loop)

    def chosen_type(self, userid, choice, popupid=None):
        '''Called when the winner makes a selection'''

        # Cancel delay
        cancelDelayed('gg_winner_menu')

        # Unload the previous gametype
        self.unload_gametype()

        # Save choice as winner_menu.gametype
        self.gametype = choice

        # Was this choice made by a player?
        if not userid in (None, False):

            # Tell everyone the next gametype
            msg('#all', 'Winner Choice Message',
                {'player': getplayername(userid),
                 'type': self.gametype})

        # Should the random message be sent?
        elif userid is None:

            # Tell everyone the next gametype
            msg('#all', 'Random Choice Message', {'type': self.gametype})

    def choose_random_gametype(self, userid=None):
        '''Chooses the next gametype randomly'''

        # Get the gametypes to choose from
        types_to_choose = self.get_gametypes()

        # Choose a random gametype
        choice = rchoice(types_to_choose)

        # Submit the choice
        self.chosen_type(userid, choice)

    def get_gametypes(self):
        '''Returns the gametypes to be chosen from'''

        # Store the GameTypes
        types_to_choose = list(self.gametypes)

        # Get the number of players on the server
        players = getplayercount()

        # Filter out any GameType requiring
        # more players than currently on server
        types_to_choose = [gametype for gametype in types_to_choose if
            int(self.gametypes[gametype]['min_players']) <= players]

        # Are there any GameTypes remaining?
        if not types_to_choose:

            # Add all GameTypes back in
            types_to_choose = list(self.gametypes)

        # Are there more than 9 GameTypes available?
        if len(types_to_choose) > 9:

            # Make sure only 9 GameTypes are in menu
            types_to_choose = sample(types_to_choose, 9)

        # Does the list need randomized?
        elif int(ServerVar('gg_winner_menu_random')):

            # Randomize the list
            shuffle(types_to_choose)

        # Return the gametypes to choose from
        return types_to_choose

    @staticmethod
    def get_random_player(team=None):
        '''Returns a random player (from the given team if one is given)'''

        # Get a list of all userids
        useridlist = getUseridList()

        # Are bots allowed?
        if not int(allow_bots):

            # If not, filter them out
            useridlist = [userid for userid in useridlist if not isbot(userid)]

        # Are there any players on the server?
        if not useridlist:

            # If not, simply return None
            return None

        # Was a team given?
        if not team is None:

            # Get a list of players on the given team
            teamlist = [userid for userid in
                useridlist if getplayerteam(userid) == team]

            # Are there any players on the given team?
            if teamlist:

                # Return a random player from the given team
                return rchoice(teamlist)

        # If nothing else, return any random userid
        return rchoice(useridlist)

# Get the _WinnerMenu instance
winner_menu = _WinnerMenu()


# ============================================================================
# >> LOAD & UNLOAD
# ============================================================================
def load():
    '''Called when the script is first loaded'''

    # Should the default gametype be random?
    if int(ServerVar('gg_winner_menu_default_random')):

        # Choose a random default gametype
        winner_menu.choose_random_gametype(False)

    # Otherwise
    else:

        # Get the default gametype
        winner_menu.load_gametype(str(ServerVar('gg_winner_menu_default')))

    # Reset the main attributes
    winner_menu.match_reset()


def unload():
    '''Called when the script is unloaded'''

    # Unload the current gametype
    winner_menu.unload_gametype()

    # Does the menu exist?
    if exists('gg_winner_menu'):

        # Delete the menu
        delete('gg_winner_menu')


# ============================================================================
# >> GAME EVENTS
# ============================================================================
def es_map_start(event_var):
    '''Called every time a map is loaded'''

    # Load the current gametype
    winner_menu.load_gametype()


def gg_start(event_var):
    '''Called when a GunGame match starts'''

    # Reset the main attributes
    winner_menu.match_reset()


def round_start(event_var):
    '''Called when a round starts'''

    # Reset the round attributes
    winner_menu.round_reset()


def round_end(event_var):
    '''Called when a round ends'''

    # Get the player that won the round
    winner_menu.round_end(int(event_var['winner']), int(event_var['reason']))


def bomb_exploded(event_var):
    '''Called when the bomb explodes'''

    # Add the bomber to the specials dictionary
    winner_menu.specials[0] = int(event_var['userid'])


def bomb_defused(event_var):
    '''Called when the bomb is defused'''

    # Add the defuser to the specials dictionary
    winner_menu.specials[6] = int(event_var['userid'])


def hostage_rescued(event_var):
    '''Called when a hostage is rescued'''

    # Set the rescuer as the latest rescuer in the specials dictionary
    winner_menu.specials[10] = int(event_var['userid'])


def player_death(event_var):
    '''Called when a player dies'''

    # Get the attacker
    attacker = int(event_var['attacker'])

    # Was this a suicide?
    if attacker in (int(event_var['userid']), 0):

        # If so, simply return
        return

    # Get the attacker's team
    team = int(event_var['es_attackerteam'])

    # Was this a team-kill?
    if int(event_var['es_userteam']) == team:

        # If so, simply return
        return

    # Set the attacker as the team's last killer
    winner_menu.specials[team] = attacker

    # Add 1 to the killer's total
    winner_menu.team_killers[team][attacker] += 1


# ============================================================================
# >> GUNGAME WIN EVENTS
# ============================================================================
def gg_win(event_var):
    '''Called when a player wins the match'''

    # Set the winner as the player that won the match
    winner_menu.winner = int(event_var['winner'])

    # Send the winner the menu
    winner_menu.send_winner_menu()


def gg_team_win(event_var):
    '''Called when a team wins the match'''

    # Send the winner menu
    winner_menu.send_winner_menu(int(event_var['winner']))
